import mako.exceptions
from mako.lookup import TemplateLookup

import pypfop


def get_lookup(lookup_dirs, input_enc='utf-8', output_enc='utf-8'):
    if isinstance(lookup_dirs, str):
        lookup_dirs = (lookup_dirs,)
    return TemplateLookup(directories=lookup_dirs,
                          input_encoding=input_enc,
                          output_encoding=output_enc)

class Template(object):

    def __init__(self, template_path, lookup):
        self.template_path = template_path
        self.lookup = lookup


    def render(self,  params):
        template = self.lookup.get_template(self.template_path)
        try:
            return template.render(**params)
        except Exception:
            raise Exception(mako.exceptions.text_error_template().render())


class TemplateFactory(object):
    __skel_dir__ = pypfop.skeldir_for('mako')
    
    def __init__(self, lookup_dirs=None, use_skels=True):
        lookup_dirs = self._get_lookup_dirs(lookup_dirs, use_skels)
        self.lookup = get_lookup(lookup_dirs)

    def __call__(self, template):
        return Template(template, self.lookup)

    def _get_lookup_dirs(self, lookup_dirs, use_skels):
        if lookup_dirs is None:
            lookup_dirs = ['.',]
        else:
            if isinstance(lookup_dirs, str):
                lookup_dirs = [lookup_dirs,]
            else:
                lookup_dirs =  list(lookup_dirs)
        if use_skels:
            lookup_dirs.append(self.__skel_dir__)
        return lookup_dirs



import sys
from xml.etree import ElementTree

import lxml.etree
import cssutils
import cssselect


def _translate_stylesheet(stylesheet, translator):
    for rule in stylesheet.cssRules:
        # If is an import rule, the generator is going to be recursive!
        if isinstance(rule, cssutils.css.CSSImportRule):
            for trans_rule in \
                    _translate_stylesheet(rule.styleSheet, translator):
                yield trans_rule
        elif isinstance(rule, cssutils.css.CSSStyleRule):
            selector = rule.selectorText
            xsel = translator.css_to_xpath(selector)
            yield (xsel, {prop.name: prop.value
                          for prop in \
                          rule.style.getProperties()})


def apply_css_sheets(xmlstring, *sheets):
    tree = lxml.etree.fromstring(xmlstring)
    final_rules = translate_css_to_xpath(*sheets)    
    for (xsel, attribs) in final_rules:
        for e in tree.xpath(xsel):
            for name, value in attribs.items():
                e.attrib[name] = value
    # After the styles related to the class has been inlined,
    # remove the class attribute to be a valid FO.
    for elem_with_class in tree.xpath('descendant-or-self::*[@class]'):
        del elem_with_class.attrib['class']
    return lxml.etree.tostring(tree)
    

def fofactory(tag, attribs):
    """
    Factory to create each element with the fo: namespace.
    """
    tag = foname(tag)
    return ElementTree.Element(tag, attribs)


def foname(tag):
    """Append the fo: namespace to the tag.
    If the tag already have a namespace return the tag unmodified.
    """
    # if no namespace is specified use the fo:
    if not ':' in tag:
        return 'fo:%s' % tag
    else:
        return tag


def translate_css_to_xpath(*sheets):
    gtrans = cssselect.GenericTranslator()
    xpath_n_styles = [] # using a list, to be hable to "cascade".
    for sheet_path in sheets:
        stylesheet = cssutils.parseFile(sheet_path)
        xpath_n_styles.extend(_translate_stylesheet(stylesheet, gtrans))
    return xpath_n_styles
            
           
def translate_to_fo(xmlstring, encoding):
    """Add the fo: namespace to all the objects in the xml."""
    builder = FOBuilder(fofactory) 
    fop_parser_creator = ElementTree.XMLParser(target=builder)
    fop_parser_creator.feed(xmlstring)
    foroot = fop_parser_creator.close()
    foroot.attrib['xmlns:fo'] = 'http://www.w3.org/1999/XSL/Format'
    if encoding is None:
        encoding = sys.getdefaultencoding()
    doctype = '<?xml version="1.1" encoding="%s"?>\n' % encoding
    return b''.join((doctype.encode(encoding),  ElementTree.tostring(foroot)))


def xml_to_fo_with_style(xmlstring, csssheets, encoding=None):
    if csssheets is not None:
        if isinstance(csssheets, str):
            xmlstring = apply_css_sheets(xmlstring, csssheets)
        else:  # asume it is an iterator with sheets.
            xmlstring = apply_css_sheets(xmlstring, *csssheets)
    return translate_to_fo(xmlstring, encoding)


class FOBuilder(ElementTree.TreeBuilder):

    def end(self, tag):
        tag = foname(tag)
        super().end(tag)




import os
import logging
import inspect
import itertools
import tempfile
import subprocess as subp

from pypfop.conversion import xml_to_fo_with_style

__version__ = '0.1.1'
__version_info__ = __version__.split('.')

BASEDIR =  os.path.join(os.getcwd(), os.path.dirname(__file__))
SKELDIR = os.path.join(BASEDIR, 'skeletons')
FOP_ENV_VAR = 'FOP_CMD'
VALID_OFORMATS = frozenset(('awt', 'pdf', 'mif', 'rtf',
                            'tiff', 'png', 'pcl', 'ps', 'txt'))


def skeldir_for(tname):
    return os.path.join(SKELDIR, tname)
    

class Document(object):
    """Basic document, you should define the `__template__` attribute
    on each subclass or pass it as a parameter in the __init__,
    the requirement for the template is that need to have a callable
    property which accept two arguments (properties, format) that
    will be translated by the renderer to a single string.
    """
   
    __style_sheets__ = ()
    __style_dir__ = '.'
    __defparams__ = {}
    __fop_cmd__ = ''
    __template__ = None
    __tempdir__ = tempfile.gettempdir()


    def __init__(self, template=None, stylesheets=(), oformat='pdf',
                 instparams=None, styledir=None, fop_cmd=None,
                 tempdir=None, debug=False):
        self.styledir = styledir or self.__style_dir__
        self.tempdir = tempdir or self.__tempdir__
        self._setup_logging()
        self.debug = debug
        self.template = self._check_template(template)
        self.oformat = self._check_oformat(oformat)
        self.fop_cmd = self._find_fop_cmd(fop_cmd)
        self.defparams = self._get_inst_params(instparams)
        self.ssheets = self._ssheets_with_abspath(stylesheets)

        

    def _check_template(self, template):
        # I wonder if this method really needs to be that pedantinc...
        if template is None and self.__template__ is None:
            raise Exception("Cannot build %s, " 
                            "neither __template__ or template is set" % \
                            self.__class__.__name__ )
        else:
            if hasattr(template, 'render') and callable(template.render):
                if inspect.ismethod(template.render) and \
                       len(inspect.getargspec(template.render).args) == 2:
                    return template
                elif inspect.isfunction(template.render) and \
                         len(inspect.getargspec(template.render).args) == 1:
                    return template
                else:
                    raise Exception(
                        'The template object %s does not implement '
                        'a 1 argument "render" property (method)' % template)
            else:
                raise Exception(
                    'The template object %s does not implement '
                    'a callable "render" property (method)' % template)
    

    def _check_oformat(self, oformat):
        oformat = oformat.lower()
        if oformat in VALID_OFORMATS:
            return oformat
        else:
            raise Exception('Invalid output format %s' % oformat)


    def _find_fop_cmd(self, fop_cmd):
        if fop_cmd is not None and fop_cmd:
            return fop_cmd
        if self.__fop_cmd__:
            return self.__fop_cmd__
        else:                
            try:
                return os.environ[FOP_ENV_VAR]
            except KeyError:
                raise Exception(
                    'Unable to find the path to execute FOP.'
                    'Check the environment variable "%s"' % FOP_ENV_VAR)


    def _setup_logging(self):
        """Overwrite this method in case that you want
        a more "intelligent" logging handler.
        """
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter())
        log = logging.getLogger('pypfop')
        log.setLevel(logging.DEBUG)
        log.addHandler(handler)
        self.log = log

    def _ssheets_with_abspath(self, ssheets):
        if isinstance(ssheets, str): 
            ssheets = [ssheets,]
        return [os.path.join(self.styledir, sheet)
                for sheet in itertools.chain(self.__style_sheets__, ssheets)]
            

    def _debug_msg(self, msg, label=''):
        if self.debug:  # TODO: Improve this method.
            if isinstance(msg, str):
                self.log.debug('%s: %s' % (label, msg))
            else:
                self.log.debug('%s: %s' % (label, msg.decode()) )

        
    def _get_inst_params(self, params):
        if isinstance(params, dict) and params:
            defparams = self.__defparams__.copy()
            defparams.update(params)
            return defparams
        else:
            return self.__defparams__


    def _generate_xslfo(self, params, copy_params):
        if copy_params:
            params = params.copy()
        params.update(self.defparams)
        xml = self.template.render(params)
        self._debug_msg(xml, 'Generated XML')
        xslfo = xml_to_fo_with_style(xml, self.ssheets)
        self._debug_msg(xslfo, 'Generated XSL-FO from xml_to_fo')
        return xslfo


    def _get_tempfile(self):
        fd, ofilepath = tempfile.mkstemp('.' + self.oformat, dir=self.tempdir)
        os.close(fd)
        return ofilepath


    def generate(self, params, oformat=None, copy_params=False):
        """Return the name of the generated document.
        
        Raise Exception in case of an error with the fop command and use
        the stderr of the command as the body of the Exception.
        """
        if oformat is None:
            oformat = self.oformat
        else:
            oformat = self._check_oformat(oformat)
        xslfo = self._generate_xslfo(params, copy_params)
        ofilepath = self._get_tempfile()
        cmdargs = [self.fop_cmd, '-q', '-fo', '-',  # - == stdin
                   '-%s' % oformat, ofilepath]
        self._debug_msg('cmdline %s' % cmdargs)
        with subp.Popen(cmdargs,
                        stdin=subp.PIPE, stdout=subp.PIPE,
                        stderr=subp.PIPE) as proc:
            stdout, stderr = proc.communicate(xslfo)
            self._debug_msg(stderr, 'STDERR of fop command')
            if b'SEVERE: Exception' in stderr or \
               b'GRAVE: Exception' in stderr:
                raise Exception(stderr.decode())
            else:
                return ofilepath





import os
from setuptools import setup

REPODIR = os.path.dirname(os.path.realpath(__file__))

def _get_longdesc(repodir):
    return open(os.path.join(repodir, 'README')).read()


def _get_license(repodir):
    return open(os.path.join(repodir, 'LICENSE')).read()


def _get_version(repodir):
    version = 'UNKNOWN'
    init = open(os.path.join(repodir, 'pypfop', '__init__.py'))
    for line in init.readlines():
        if '__version__' in line and '=' in line:
            version = line.split('=')[-1].strip()
            version = version.replace('"', '').replace("'", '')
            break
    init.close()
    return version


description = 'Document preprocessor for Apache FOP.'
version = _get_version(REPODIR)
longdesc = _get_longdesc(REPODIR)
license_ = _get_license(REPODIR)
classifiers = ['Development Status :: 2 - Pre-Alpha',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: Apache Software License',
               'Operating System :: OS Independent',               
               'Programming Language :: Python :: 3.0',
               'Programming Language :: Python :: 3.1',
               'Programming Language :: Python :: 3.2',
               'Programming Language :: Python :: 3.3',
               'Topic :: Office/Business',
               'Topic :: Software Development :: Libraries',
               'Topic :: Text Processing :: Filters']
requires = ['Mako',
            'cssutils',
            'cssselect',
            'lxml']
setup(name='pypfop',
      version=version,
      author='Joel Rivera',
      author_email='rivera@joel.mx',
      maintainer='Joel Rivera',
      maintainer_email='rivera@joel.mx',
      url='https://bitbucket.org/cyraxjoe/pypfop',
      provides=['pypfop',],
      packages=['pypfop',],
      classifiers=classifiers,
      install_requires=requires,
      include_package_data=True,
      platforms=['linux2', 'win32', 'cygwin', 'darwin'],
      description=description,
      long_description=longdesc)
